#ifndef  _ADSTR_
#define  _ADSTR_

/* This is the data structure for switched-capacitor circuit layout program */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change : 8/8/87 */

/* the following is the structure used for describing the various
modules in the layout, such as capacitor blocks, opamps, switches */

#include "ADdef.h"

struct _module {

    char *name;

    /* Geometric infromation about the module */

    int row, col; /* order in the x and y directions */
        /* the row number increses in the +y direction */
        /* the col number increses in the +x direction */
    int xdim, ydim; /* the x and y dimensions of the module */
    int xtrans, ytrans; /* the translations of the module */
    int transform; /* type of transfromation for the module */
    int xsep; /* spacing between this module and its left neighbor */

    struct _terminal *terminals; /* list of terminals */

    struct _sterminal *sterminals;	/* list of supply terminals */
                                        /* -hcc */

    /* particular information about each type of module */

    int type; /* CBLK, SWCH, OPAMP, CHAN, INMOD, OUTMOD */
    union {
        struct _cblk *cblk; /* information about the capacitor block */
        struct _swch *swch; /* information about the switch */
        struct _opamp *opamp; /* information about the opamp */
        struct _channel *channel; /* information about channel */
    } info;

    struct _module *next; /* next module in the ordered list of modules
                            of the same type */
    struct _module *prev; /* previous module */

    struct _lib *libptr; /* points to module's library location */
};

struct _cblk {
    struct _module *module;

    /* information used for the find height algorithm */

    int unitA; /* area of unit capacitor for this capacitor block */
    int unith; /* height of the unit capacitor for this block */
    int unitw; /* width of the unit capacitor for this block */
    int numrow; /* height of the block in # of unit capacitors */
    int numcol; /* width of the block in # of unit capacitors */

    /* in number of unit capcitors */
    int topsum, botsum, totalsum; 

    /* for each input */
    int ncontc; /* # of capacitors directly connected to the opamps */
    struct _cap *contcaps; /* list of above  capacitors */
    struct _cap *integcap; /* integrating capacitor */

    int nswitc; /* # of switched capacitors */
    struct _cap *swcaps; /* list of switched capacitors */

    struct _terminal *eqtoppt; /* equivalent top plate terminal */
    struct _terminal *eqbotpt; /* equivalent bottom plate terminal */

    int in_polarity; /* polarity of the opamp input this is connected to */
                        /* PLUS or MINUS */
    struct _opamp *opamp; /* the opamp correponding to this block */

};

struct _cap {
    struct _cblk *cblk;

    char name[40];
    int type; /* SWITCHED, CONT */

    double size; /* in farads */
    double ratio; /* in number of unit capacitors */
    int cratio; /* ceiling value of ratio */

    struct _terminal *topplate; /* the topplate terminal of the capacitor */
    struct _terminal *botplate;

    int nuc_placed; /* flag specifying nonunitcapacitors are placed */

    double tplsval, bplsval; /* sort values for topplate and botom plate */

    struct _cap *next; /* next in the ordered list of capacitors for its block*/
};


struct _swch {
    struct _module *module;

    char name[40];
    int type; /* SWIN, SWOUT2, or SWOUT3 */

    struct _terminal *amp1term; /* when clock1 high, capterm connects to this */
    struct _terminal *clk1term; /* the first controlling clock */
    struct _terminal *capterm; /* connection to capacitor terminal */
    struct _terminal *clk2term; /* the controlling clock */
    struct _terminal *amp2term; /* when cloc2 high, capterm connects to this */

    struct _terminal *clk1bart; /* complement of clk1 if it is needed */
    struct _terminal *clk2bart;

    struct _swch *nextsw; /* next switch belonging to the list of same clock*/
};

struct _opamp {
    struct _module *module;

    char name[40];

    struct _terminal *out[MAXNUMIN]; /* controlled nodes */
    struct _terminal *inp[MAXNUMIN]; /* controlling nodes */

    struct _cblk *cblk[MAXNUMIN]; /* the capacitor blocks
                                    corresponding to this opamp */
};

struct _terminal {
    struct _module *module; /* the module terminal belongs to */
    struct _cap *capacitor; /* not null only if belongs to one capacitor */
    struct _net *netptr; /* the net this terminal belongs to */
    struct _channel *channel; /* the channel terminal belongs to */

    char name[40];

    struct _eqterm *eqterms; /* list of equivalent locations */

    char *layer; /* the layer of the terminal */

    int type; /* INPUT, OUTPUT, INOUT */

    int xcoord; /* xcoordinate of the pin in its module */
    int ycoord; /* ycoordinate of the pin in its module */

    int chrlayer; /* the layer name as it appers in channel router */
    int chloc; /* TOP, BOTTOM, RIGHT, or LEFT of its channel */
    int xoffset; /* is non zero only if two pins are too close */
    int chxcoord; /* normalized actual xcoordinate of the pin */
    int chycoord; /* normalized actual ycoordinate of the pin */

    /* the two x extremes of the terminal's net in the terminal's channel */
    /* it convers only the terminals in the modules which have defined loc */
    int net_lowxlim; /* used in permutation stage */
    int net_hixlim;

    struct _terminal *next; /* next terminal in the list for its module */
    struct _terminal *nxtnett; /* next on the list of terminals for its net */
    struct _terminal *nxtcht; /* list of terminals for its channel */
};

/*
 * Information on the supply terminals. 
 * -hcc
 */
struct _sterminal {
    char	name[40];		/* Name of the terminal, e.g. Vdd.l */
    char	net[NETNAME_SIZE];	/* Name of the net, e.g. Vdd        */
    char	property[PROPERTYNAME_SIZE];
                                        /* SUPPLY or GROUND                 */
    int		side;			/* TOP, BOTTOM, RIGHT, or LEFT      */ 
    struct _sterminal *next;		/* Next one for the module          */
}; /* _sterminal */

struct _eqterm {
    struct _module *module;

    int xcoord;
    int ycoord;

    struct _eqterm *next; /* next for the list for its terminal */
};

struct _net {
    char name[40]; /* name as read from input */
    /* net number is between 1 and Numnet */
    int number; /* number assigned to the net */
    int type; /* IN, includes input of an opamp */
              /* OUT, includes output of an opamp */
              /* SWITCHED includes capterms of some switches */
              /* CLOCK, if it is controlling clock line */
              /* GROUND, the ground net */
    struct _module *module; /* not null if it belongs to only one module */
                            /* true for INPUT and OUTPUT */

    struct _terminal *firstnett; /* first terminal of this net */

    int sensitive; /* YES or NO */

    struct _terminal *chminxt[MAXNUMCH]; /* terminals belonging to this net
                                with minumum xcoordinate in its channel */
    struct _terminal *chmaxxt[MAXNUMCH]; 

    struct _net *next; /* next net on the list */
};

struct _channel {
    struct _module *module; /* the module this channel belongs to */

    int name; /* TOP , BOTTOM, or CLKCH */

    int maxdensity; /* maximum density of the channel */

    int lowchx, lowchy, hichx, hichy; /* channel boundaries in channel coords */
    int numnets; /* number of nets in the channel */

    /*** the four directions are: TOP, BOTTOM, RIGHT, LEFT */
    int numterms[MAXNUMDIR]; /* number of terminals for each direction */
    struct _terminal *tlist[MAXNUMDIR]; /* list of terminals in each of 4 dir */

    struct _xgrids *xgrids; /* list of grid lines for this channel */

    /*** information extracted from output of channel router */
    int numvias; /* number of vias belonging to this channel */
    struct _via *vialist; /* list of vias belonging to this channel */

    int numwires; /* number of wires belonging to this channel */
    struct _subnet *snetlist; /* list of subnets */

    struct _ch_matrix **ch_matrix; /* matrix of grid points */


    /*** shielding information for the channel ***/
    int numshpins; /* total number of shielding terminals */
    struct _shterm *shpins; /* list of shielding pins */ 

    int numshgr; /* total number of ground terminals for shielding */
    int *isolated_grs; /* list of flags for ground terminals */
                        /* is YES if the corresponding ground is isolated */
};

struct _xgrids {
    int xcoord; /* actual coordinate of the grid line */

    struct _terminal *topterm; 
    struct _terminal *botterm; 
};

struct _ch_matrix {
    struct _via *via; /* the via on this grid point */

    struct _snet_stack *snet_stack; /* list of subnets crossing here*/
};

struct _via {
    int net; /* the net number as appears in the channel router */
    int chx, chy; /* channel coordinates of the center */
    int layer1, layer2; /* the layers it connects to */
};

struct _subnet {
    int net;

    int chxl, chyl; /* coordinate of the low end point */
    int chxh, chyh; /* coordinate of the hight end point */

    int layer;

    int type; /* HORIZ or VERT */
};

struct _shterm {
    int shieldnet; /* the assigned net number in the shield */

    int xcoord; /* normalized actual x coordinate of the shield terminal */
    int ycoord;
};

struct _lib {
    /* pointer to library location in the file system */
    char libname[100]; /* name of the library entry */
    struct _mod_stack *stptr; /* list of modules having same library name */

    struct _bus *buslist; /* list of horizontal busses */

    struct _lib *next; /* next library entry */
};

struct _bus {
    char name[40]; /* name of the bus */

    char layer[40];
    int lowlefty; /* lowerleft y coordinate of the bus in the library */
    int uprighty;

    struct _bus *next; /* in the list for its library */
};

/* used for stack operations on module */
struct _mod_stack {
    struct _module *module; 
    struct _mod_stack *next;
};

/* used for stack operations on terminal */
struct _term_stack {
    struct _terminal *terminal;
    struct _term_stack *next;
};

/* used for stack operations on subnet */
struct _snet_stack {
    struct _subnet *subnet;
    struct _snet_stack *next;
};

#endif
