/* permute a set of capacitor blocks in an array, for Adore */
/* Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli */
/* Last Change: April 1992  -  Enrico Malavasi */

#include <stdio.h>
#include "ADglobals.h"
#include "PERMglobals.h"

/* This file contains the files for generation of state sapce
graph and the branch and bound technique to minimize the density
for abstraction of capacitor blocks.  The algorithm is based 
on the paper "Suboptimum Solution of the Back-Board Ordering with
Channel Capacity Constraint" by S. Goto et al, in trans. on CAS,
Nov. 1977, p. 645 */

static set_t *leftset, *rightset, *tmpset;

struct _permobj *permlist; /* list of permutable switches */
struct _ssgnode **openssglist; /* ssgnodes in the open list */
struct _ssgnode **binssglist; /* binary search tree for ssgnodes */


/* find the transformation of capacitor blocks */
static void
  PERMtransform()
{
    struct _module *tmpmod;

    /*** for now the capacitor blocks are alternatively transformed */
    tmpmod = modlist[CBLK];
    for ( ; tmpmod != NULL; tmpmod = tmpmod->next ) {
        if ( (tmpmod->col % 2) == 1 ) {
            tmpmod->transform = YES;
        } else {
            tmpmod->transform = NO;
        }
    }
}

PERMcblk(epsilon, numsrch_l)
double epsilon;
int numsrch_l; /* number of levels to be searched */
{
    PERMcapNets();

    PERMcblkSsgInit();

    PERMcblkEpsAlgo(epsilon, numsrch_l);

    PERMfindXsep();

    PERMfindXtrans ( CBLK );

    /* find the transformation of capacitor blocks */
    PERMtransform();
}

/* inititalize the state space graph for capacitor blocks */
/* fill each permobj with the information from its corresponding cblk */
PERMcblkSsgInit( ) 
{
    int counter;

    set_t *set_new();
    struct _module *tmpmod;


    CALLOC(permlist, Numcblk + 1, struct _permobj);
    for ( tmpmod = modlist[CBLK]; tmpmod != NULL; tmpmod = tmpmod->next ) {
        permlist[tmpmod->col].module = tmpmod;
    }

    CALLOC(openssglist, Numcblk + 1, struct _ssgnode *);
    CALLOC(binssglist, Numcblk + 1, struct _ssgnode *);
    for ( counter = 0; counter <= Numcblk; counter++ ) {
        openssglist[counter] = binssglist[counter] = NULL;
    }

    leftset = set_new(numnewnets);
    rightset = set_new(numnewnets);
    tmpset = set_new(numnewnets);
}

/* the main algorithm for finding the order */
PERMcblkEpsAlgo(epsilon, numsrch_l)
double epsilon;
int numsrch_l;
{
    struct _ssgnode *startnode;
    struct _ssgnode *newnode, *PERMnextSsgnode();
    int maxlevel; /* maximum level used for search */
    int numexp;

    /*  initialize vs and put it in the open list */
    CALLOC(startnode, 1, struct _ssgnode);
    startnode->lastobj = NULL;
    startnode->level = maxlevel = 0;
    startnode->parent = NULL;
    startnode->next = NULL;
    startnode->setrep = set_new(Numcblk + 2);
    /* the INMOD is the first block */
    set_insert(startnode->setrep, 0);
    startnode->whichset = OPEN;

    startnode->cost = PERMssgNodeCost(startnode->setrep);
    startnode->hfunc = 0;
    startnode->est_cost = startnode->cost;

    /* add to the open list */
    PERMaddSsglist(startnode);

    numexp = 0;
    while ( 1 ) {
        /***  find the closest vertex to the goal node satisfying
         mincost <= cost(v) <= (1 + epsilon) * mincost */

        if ( (newnode = PERMnextSsgnode(maxlevel, numsrch_l, epsilon))==NULL ){
            printf("eps_algo: the open list is empty\n");
            exit(-1);
        }

        /*  if we have reached the goal node, find the solution path
        by following the back pointers */
        else if ( newnode->level == Numcblk ) {
            /* we have reached the goal vertex */
            PERMgenNewPerm( CBLK, newnode );
            return( YES );
        }

        /* now expand the new vertex */
        PERMexpandSsg( newnode );
        numexp++;

        if ( (newnode->level + 1 ) > maxlevel ) {
            maxlevel = newnode->level + 1;
        }
    }
}
 
/* re order the list pointed to by modlist[MODTYPE] */
PERMgenNewPerm( MODTYPE, ssgnode)
int MODTYPE; /* SWCH CBLK or OPAMP */
struct _ssgnode *ssgnode;
{
    struct _ssgnode *tempssg;
    struct _module  *tmpmod, *lastmod;
    int counter;

    tempssg = ssgnode;
    lastmod = NULL;
    for ( ; tempssg->parent != NULL; tempssg = tempssg->parent ) {
        tempssg->lastobj->module->next = lastmod;
        lastmod = tempssg->lastobj->module;


		/* The following lines need to be added if ssg graph is
			used for ordering switches and op-amps */
		/*
        if ( MODTYPE == OPAMP || MODTYPE == SWCH ) {
            PERMssgEcost(tempssg->lastobj, tempssg->xtrans,
                                &tempssg->xsep, &tempssg->transform);
            lastmod->xtrans = tempssg->xtrans;
            lastmod->transform = tempssg->transform;
            lastmod->xsep = tempssg->xsep;
        }
		*/

    }

    if ( tempssg->level != 0 ) {
        printf("PERMgenNewPerm: wrong back pointers\n");
        exit(-1);
    }

    modlist[MODTYPE] = lastmod;

    counter = 0;
    for ( tmpmod = modlist[MODTYPE]; tmpmod != NULL; tmpmod = tmpmod->next ) {
        counter++;
        tmpmod->col = counter;
    }
}



/* expand one node of state space graph */
/* returns NO if ssgnode is not expanded */
PERMexpandSsg( ssgnode )
struct _ssgnode *ssgnode;
{
    struct _ssgnode *tempssg, *newssg, *PERMis_in_ssglist();
    set_t *tempset, *set_new_copy();
    struct _permobj *tempobj;
    int inset; /* is in the set or not */
    int chorder; /* the order of the child in the permlist ordering */
    int tempcost; 
    int level; /* level and xtranslation of children of this ssg node */
    int temp_est_cost;


    /* level of the ssgnode's children */
    level = ssgnode->level + 1;

    /* find if a given object is in the set defined by this ssgnode */
    for ( chorder = 1; chorder <= Numcblk; chorder++) {
        if (is_in_set(ssgnode->setrep, chorder)) {
            inset = YES;
        }
        else {
            inset = NO;
        }
        if ( inset == NO ) {
            /* copy the parent set */
            tempset = set_new_copy(ssgnode->setrep);
            /* set the bit corresponding to chorder */
            set_insert(tempset, chorder);

            tempobj = permlist + chorder;
            temp_est_cost = PERMssgNodeCost( tempset );


            if ( (tempssg = PERMis_in_ssglist(tempset, level)) == NULL ) {
                /* it is not in the closed or open lists */
                /* generate the child */
                CALLOC(newssg, 1, struct _ssgnode);
                newssg->setrep = tempset;
                newssg->level = level;
                newssg->lastobj = tempobj;
                newssg->parent = ssgnode;
                newssg->est_cost = newssg->cost = temp_est_cost;
                newssg->whichset = OPEN;
                PERMaddSsglist(newssg, openssglist);
            } else if ( tempssg->whichset==OPEN &&
                                      temp_est_cost < tempssg->est_cost ) {
                /* change parents */
                if (tempssg->level != level) {
                    printf("PERMexpandSsg:levels don't match\n");
                    exit(-1);
                }
                tempssg->lastobj = tempobj;
                tempssg->parent = ssgnode;
                tempssg->cost = temp_est_cost;
                tempssg->est_cost = temp_est_cost;
                PERMupdatOpenlist(tempssg);
            }
        }
    }
    return( YES );
}

/* alternative approach to finding density at the end of each block */
/*** an active set of nets for a given block is the set of those
nets which extend beyond a given block, their number is the density
at the end of that block */
/*** for every net connected to this block find whether it extends
beyond this block, then add it to the set of active nets of this block */ 
/*** for every net in the active set of its parent find if it 
terminates in this block otherwise add to active set for this block */

/* evaluate the cost of a ssgnode */
PERMssgNodeCost(ssgnode_set)
set_t *ssgnode_set;
{
    int chorder;


    set_clear(leftset);
    set_clear(rightset);

    for ( chorder = 0 ; chorder <= Numcblk + 1; chorder++ ) {
        if ( is_in_set(ssgnode_set, chorder) ) {
            set_copy(tmpset, rightset);
            set_or(rightset, net_matrix[chorder], tmpset);
        }
        else {
            set_copy(tmpset, leftset);
            set_or(leftset, net_matrix[chorder], tmpset);
        }
    }
    set_and( tmpset, rightset, leftset);
    /*** cardinality of the intersection */
    return( set_ord( tmpset ) );
}



/* add a ssgnode to binary search tree and open list */
PERMaddSsglist(ssgnode)
struct _ssgnode *ssgnode;
{
    struct _ssgnode *prevssg;
    struct _ssgnode **ssglist;
    int keyval;

    /* sort according to cost use next pointers */
    ssglist = openssglist;
    prevssg = ssglist[ssgnode->level];

    if ( prevssg == NULL || ssgnode->est_cost <= prevssg->est_cost ) {
        ssgnode->next = ssglist[ssgnode->level];
        ssglist[ssgnode->level] = ssgnode;
    }
    else {
        while ( prevssg->next != NULL && prevssg->next->est_cost 
                                        < ssgnode->est_cost ) {
            prevssg = prevssg->next;
        }
        ssgnode->next = prevssg->next;
        prevssg->next = ssgnode;
    }


    /* place according to set representation in the binary tree */
    ssglist = binssglist;
    prevssg = ssglist[ssgnode->level];

    if ( prevssg == NULL ) {
        ssglist[ssgnode->level] = ssgnode;
        ssgnode->leftson = ssgnode->rightson = NULL;
    }
    else {
        while ( prevssg != NULL ) {
            if ((keyval = set_compare(prevssg->setrep, ssgnode->setrep))== 0 ) {
                printf("add_closedlist: the set already in the list\n");
                exit(-1);
            }
            else if ( keyval < 0 ) {
                if (  prevssg->leftson == NULL ) {
                    prevssg->leftson = ssgnode;
                    ssgnode->leftson = ssgnode->rightson = NULL;
                    break;
                }
                else {
                    prevssg = prevssg->leftson;
                }
            }
            else if ( keyval > 0 ) {
                if ( prevssg->rightson == NULL ) {
                    prevssg->rightson = ssgnode;
                    ssgnode->leftson = ssgnode->rightson = NULL;
                    break;
                }
                else {
                    prevssg = prevssg->rightson;
                }
            }
        }
    }
}
                

/* find whether the node is in binary search tree */

struct _ssgnode *PERMis_in_ssglist(compset, level)
set_t *compset; /* the set to be found */
int level; /* level at which the set could be found */
{
    struct _ssgnode *tempssg;
    struct _ssgnode **ssglist;
    int keyval;

    ssglist = binssglist;

    tempssg = ssglist[level];
    if ( tempssg == NULL ) {
        return ( NULL ) ;
    }
    while ( tempssg != NULL ) {
        if ( (keyval = set_compare(tempssg->setrep, compset)) == 0 ) {
            return( tempssg );
        }
        else if ( keyval < 0 ) {
            if ( tempssg->leftson == NULL ) {
                return ( NULL );
            }
            else {
                tempssg = tempssg->leftson;
            }
        }
        else if ( keyval > 0 ) {
            if ( tempssg->rightson == NULL ) {
                return ( NULL );
            }
            else {
                tempssg = tempssg->rightson;
            }
        }
    }
    printf("PERMis_in_ssglist: error\n");
    exit(-1);
}



/* when a shorter path is found for a given node, update the location
of this node in its corresponding level */

PERMupdatOpenlist( ssgnode )
struct _ssgnode *ssgnode;
{
    struct _ssgnode *tempssg, *prevssg;

    if ( ssgnode->whichset == OPEN ) {
        prevssg = openssglist[ssgnode->level];
        if ( prevssg == NULL ) {
            printf("PERMupdatOpenlist: the level in openlist is empty\n");
            exit(-1);
        }
        else if ( prevssg == ssgnode ) {
            /* remove from open list */
            openssglist[ssgnode->level] = ssgnode->next;
            ssgnode->next = NULL;
        }
        else {
            while ( prevssg->next != NULL && prevssg->next != ssgnode ) {
                prevssg = prevssg->next;
            }
            if ( prevssg->next == NULL ) {
                printf("PERMupdatOpenlist: ssgnode is not in open list\n");
                exit(-1);
            }
            /* remove from open list */
            prevssg->next = ssgnode->next;
            ssgnode->next = NULL;
        }
    }
    else if ( ssgnode->whichset == CLOSED ) {
        ssgnode->whichset = OPEN;
    }

    /* now find the order with the new key value */
    /* reinsert in the open list */
    prevssg = openssglist[ssgnode->level];

    if ( prevssg == NULL || ssgnode->est_cost <= prevssg->est_cost ) {
        ssgnode->next = openssglist[ssgnode->level];
        openssglist[ssgnode->level] = ssgnode;
    }
    else {
        while ( prevssg->next != NULL && prevssg->next->est_cost <
                                            ssgnode->est_cost ) {
            prevssg = prevssg->next;
        }
        ssgnode->next = prevssg->next;
        prevssg->next = ssgnode;
    }
}

/* find the next node to be expanded from the open list */
/* find the node with minimum cost, then find the node v with largest
level number satisfying: mincost <= cost(v) <= (1+ eps) * mincost */

struct _ssgnode *PERMnextSsgnode(maxlevel, numsrch_l, epsilon)
int maxlevel; /* the maximim level to be searched */
int numsrch_l; /* number of levels to be searched */
double epsilon;
{
    int level;
    int minlevel;
    double mincost;
    struct _ssgnode *tempssg;

    mincost = 1.e99;
    /*
    minlevel = maxlevel - numsrch_l + 1;
    if ( minlevel < 0 ) {
        minlevel = 0;
    }
    */
    minlevel = 0;
    for (level = minlevel; level <= maxlevel; level++) {
        tempssg = openssglist[level];
        if ( tempssg != NULL && tempssg->est_cost < mincost ) {
            mincost = tempssg->est_cost;
        }
    }

    for ( level = maxlevel; level >= minlevel; level-- ) {
        tempssg = openssglist[level];
        if ( tempssg != NULL &&
                     (double)tempssg->est_cost <= ((1 + epsilon) * mincost) ) {
            break;
        }
    }

    if ( tempssg == NULL ) {
        return(NULL);
    }
    /* now remove from open list and add to closed list */
    openssglist[level] = tempssg->next;
    tempssg->next = NULL;

    tempssg->whichset = CLOSED;

    return tempssg;
}
