/*  Permute a set of modules in an array, for Adore2 
 *  This file contains the subroutines for simulated-annealing
 *  one-dimensioanl placement of op-amps and switches
 *
 * Author: Hormoz Yaghutiel, Advisor: Alberto Sangiovanni-Vincentelli
 * Last update:  March 1992  -  Enrico Malavasi */

#include <stdio.h>
#include <math.h>
#include "PERMglobals.h"
#include "TEKtech.h"

#define SWAP 1
#define DISPLACE 2
#define MIRROR 3

static int totxdim;    /* total length of all modules with no seperation */
static int maxxdim;    /* total x dimension of the layout */
static int nummod;     /* total number of given module */
static int numcostinc; /* number of moves which increase cost */

static double seed;       /* needs to be initialized to .314... for evey SA algo */
static double newoverlap;
static double oldoverlap; /* total amount of overlap */

static void
  PERMsortTerms(lastobj, transform)
struct _permobj *lastobj;
int transform; /* YES or NO */
{
    int gap, icount, jcount;
    int temp;
    int nump;

    /* the sorting algorithm is from p 58 of C by Kernighan & Ritchie */
    nump = 2 * lastobj->numterms;

    /* sort lastobj->limitlist[transform][0] ... 
            lastobj->limitlist[transform][nump -1] into increasing order */
    for ( gap = nump/2; gap > 0 ; gap /= 2) {
        for ( icount = gap; icount < nump; icount++ )  {
            for ( jcount = icount-gap; jcount>=0 &&
                     lastobj->limitlist[transform][jcount] >
                     lastobj->limitlist[transform][jcount + gap]; jcount-=gap) {
                temp = lastobj->limitlist[transform][jcount];
                lastobj->limitlist[transform][jcount] = 
                            lastobj->limitlist[transform][jcount + gap];
                lastobj->limitlist[transform][jcount + gap]  = temp;
            }
        }
    }

    /* evaluate cost at breakpoints */
	/* cost in the flat part is the distance of the middle break points */
    lastobj->fvalue[transform][nump / 2] =
                    lastobj->limitlist[transform][nump/2] -
                    lastobj->limitlist[transform][nump/2 - 1];
    for ( icount = nump / 2 + 1; icount <= nump - 1; icount++) {
        lastobj->fvalue[transform][icount] = 
                lastobj->fvalue[transform][icount - 1] +
                (icount - nump/2) * (lastobj->limitlist[transform][icount] -
                    lastobj->limitlist[transform][icount - 1]);
    }
    lastobj->fvalue[transform][nump / 2 - 1] = 
    				lastobj->fvalue[transform][nump / 2];
    for ( icount = nump / 2 - 2; icount >= 0; icount--) {
        lastobj->fvalue[transform][icount] = 
                lastobj->fvalue[transform][icount + 1] +
                (nump/2 - 1 - icount) * 
                (lastobj->limitlist[transform][icount + 1] -
                    lastobj->limitlist[transform][icount]);
    }
}

static void
  PERMmSsgInit( MODTYPE, flag  ) 
/* This has to be done after block dimension evaluation
 * fill each permobj with the information from its corresponding switch */
int MODTYPE;   /*  It can only be SWCH or OPAMP */
int flag; /* EXACT or APPROX */
{
    int counter;
    int modcnt; /* counter for number of switches */
    int numterms; /* counter for number of terminals */
    int xsep, transform; /* dummy variables */

    struct _module *tmpmod, *prevmod;
    struct _permobj *lastobj;
    struct _terminal *tmpterm;

    if((MODTYPE != SWCH) && (MODTYPE != OPAMP)) {
	errRaise("PERMmSsgInit;1: internal error", -1);
    }
    nummod = ( MODTYPE == SWCH ) ? Numsw : Numinteg;

    CALLOC(permlist, nummod + 1, struct _permobj);

    maxxdim = modlist[OUTMOD]->xtrans + modlist[OUTMOD]->xsep;

    modcnt = 0;
    totxdim = 0;
    prevmod = NULL;
    for (tmpmod = modlist[MODTYPE]; tmpmod != NULL; tmpmod = tmpmod->next) {
        modcnt++;
        totxdim += tmpmod->xdim;
        permlist[modcnt].module = tmpmod;
	/* prevmod points to the previous module, needed for swapping */
	permlist[modcnt].module->prev = prevmod;
	prevmod = tmpmod;
	
        numterms = permlist[modcnt].numterms = 0;
        for ( tmpterm = tmpmod->terminals; tmpterm != NULL;
	     tmpterm = tmpterm->next ) {
	    
            /*** evaluate the lowx and highx extermes of the net
	      containing this terminal ( only previously placed 
	      terminals are considered when flag == EXACT and previosly
	      placed blocks considered when flag == APPROX) */
	    
            if ( PERMnetLimits( tmpterm, flag ) != GROUND ) {
                permlist[modcnt].numterms += 1;
		numterms = permlist[modcnt].numterms;
                permlist[modcnt].limitlist[NO][ 2 * numterms - 2 ] = 
		  tmpterm->net_lowxlim - tmpterm->xcoord;
                permlist[modcnt].limitlist[NO][ 2 * numterms - 1 ] = 
		  tmpterm->net_hixlim - tmpterm->xcoord;
                permlist[modcnt].limitlist[YES][ 2 * numterms - 2 ] = 
		  tmpterm->net_lowxlim - (tmpmod->xdim - tmpterm->xcoord);
                permlist[modcnt].limitlist[YES][ 2 * numterms - 1 ] = 
		  tmpterm->net_hixlim - (tmpmod->xdim - tmpterm->xcoord );
            }
        }
	
	
        /* for this permobj, the smallest x which minimizes cost */
        /* is between n and n+1 elements of the sorted list */
        /* there are 2 * numpins elements in limit list, 
	   sort them in the increasing order */
        PERMsortTerms(permlist + modcnt,NO);
        PERMsortTerms(permlist + modcnt,YES);
    }
    
    /* generate the initial state only the first time ever */
    if ( flag == APPROX ) {
        for ( modcnt = 1; modcnt <= nummod; modcnt++ ) {
            tmpmod = permlist[modcnt].module;
            tmpmod->transform = NO;
            tmpmod->xsep = Round_Int((maxxdim - totxdim) / nummod,
                                     resolution, FLR);
            tmpmod->xtrans = 0;
        }
        PERMfindXtrans( MODTYPE );
    }
}

static double
  PERMmakeRand()
/* Create a random number  in the interval 0 - 1.
 * The global variable seed should be initialized with 0.31415...
 * once before using the function.
 * For each run of the program using the same initialization you get
 * the same sequence of random numbers. */
{
    while( 1 ) {
	seed = seed * seed;
	if ( seed < .3 ) {
	    seed += .4;
	    return ((double) (seed * 100000. - (double)((int)(seed * 100000.))+5e-8));
	}
    }
}

static int
  PERMchooseRand( numstates )
int numstates;
/* Generates an integer between 1 and numstates  */
{
    int temprand;

    while( (temprand=(int)( ceil(numstates * PERMmakeRand()))) == 0 )   ;

    if (temprand > numstates)    temprand = numstates;
    if ( temprand < 1 )          temprand = 1;

    return temprand;
}

static void
  PERMgenMove(movetype, object1, object2, displacement)
int *movetype;
int *object1, *object2;
int *displacement;
/* This procedure generates the move, but the actual move is
 * handled through another program                         */
{
    int rand0;
    int numpositions; /* number of slots a module can move in */
    int randpos;
    struct _module *tmpmod;

    if( nummod < 2 ) {
	*movetype = DISPLACE;
	*object1 = 1;
	*object2 = 0;
	*displacement = 0;
	return;
    }

    *object1 = PERMchooseRand(nummod);
    if ((rand0 = PERMchooseRand(1000)) <= 350) {
        *movetype = SWAP;
        while( (*object2=PERMchooseRand(nummod)) == *object1)  ;

    } else if ( rand0 <= 700 ) {
        *movetype = DISPLACE;
        *object2 = 0;

	tmpmod = permlist[*object1].module;
	if ( tmpmod->next != NULL ) {
	    numpositions = (max(0,tmpmod->xsep) + max(0,tmpmod->next->xsep)) / resolution;
	} else {
	    numpositions = (max(0, maxxdim - tmpmod->xtrans - tmpmod->xdim -
				tmpmod->xsep) + max(0,tmpmod->xsep)) / resolution;
	}
	/* randpos is between 0 and numpositions */
	if ( numpositions == 0 ) {
	    *displacement = 0;
	} else {
	    randpos = PERMchooseRand( numpositions + 1 ) - 1;
	    *displacement = randpos * resolution - max(0,tmpmod->xsep);
	}
    } else {
        *movetype = MIRROR;
        *object2 = 0;
    }
}    

static void
  PERMmakeMove(movetype, object1, object2, displacement, MODTYPE)
int movetype, object1, object2;
int displacement;
int MODTYPE;
{
    int counter;

    int tmpxtrans, tmpxsep, tmpcol;
    int deltaxdim; /* used when two different width modules are swpaed */
    struct _permobj *tmpobj1, *tmpobj2;
    struct _module *tmpmod;

    tmpobj1 = permlist + object1;
    tmpobj1->cost = tmpobj1->tmpcost;
    if ( movetype == SWAP ) {
	tmpobj2 = permlist + object2;
	
	tmpxsep = tmpobj2->module->xsep;
	tmpobj2->module->xsep = tmpobj1->module->xsep;
	tmpobj1->module->xsep = tmpxsep;
	
	tmpxtrans = tmpobj2->module->xtrans;
	tmpobj2->module->xtrans = tmpobj1->module->xtrans;
	tmpobj1->module->xtrans = tmpxtrans;
	
	tmpcol = tmpobj2->module->col;
	tmpobj2->module->col = tmpobj1->module->col;
	tmpobj1->module->col = tmpcol;
	
	if ( tmpobj2->module->next == tmpobj1->module ) {
	    if ( tmpobj2->module->prev != NULL ) {
		tmpobj2->module->prev->next = tmpobj1->module;
	    } else { modlist[MODTYPE] = tmpobj1->module; }
	    
	    if ( tmpobj1->module->next != NULL ) {
		tmpobj1->module->next->prev = tmpobj2->module;
	    }
	    
	    tmpobj1->module->prev = tmpobj2->module->prev;
	    tmpobj2->module->prev = tmpobj1->module;
	    
	    tmpobj2->module->next = tmpobj1->module->next;
	    tmpobj1->module->next = tmpobj2->module;
	} else if ( tmpobj1->module->next == tmpobj2->module ) {
	    if ( tmpobj1->module->prev != NULL ) {
		tmpobj1->module->prev->next = tmpobj2->module;
	    } else { modlist[MODTYPE] = tmpobj2->module; }
	    
	    if ( tmpobj2->module->next != NULL ) {
		tmpobj2->module->next->prev = tmpobj1->module;
	    }
	    
	    tmpobj2->module->prev = tmpobj1->module->prev;
	    tmpobj1->module->prev = tmpobj2->module;
	    
	    tmpobj1->module->next = tmpobj2->module->next;
	    tmpobj2->module->next = tmpobj1->module;
	} else {
	    if ( tmpobj2->module->prev != NULL ) {
		tmpobj2->module->prev->next = tmpobj1->module;
	    }
	    else { modlist[MODTYPE] = tmpobj1->module; }
	    
	    if ( tmpobj1->module->prev != NULL ) {
		tmpobj1->module->prev->next = tmpobj2->module;
	    } else { modlist[MODTYPE] = tmpobj2->module; }
	    
	    if ( tmpobj2->module->next != NULL ) {
		tmpobj2->module->next->prev = tmpobj1->module;
	    }
	    if ( tmpobj1->module->next != NULL ) {
		tmpobj1->module->next->prev = tmpobj2->module;
	    }
	    
	    tmpmod = tmpobj2->module->prev;
	    tmpobj2->module->prev = tmpobj1->module->prev;
	    tmpobj1->module->prev = tmpmod;
	    
	    tmpmod = tmpobj2->module->next;
	    tmpobj2->module->next = tmpobj1->module->next;
	    tmpobj1->module->next = tmpmod;
	}
        
	tmpobj2->cost = tmpobj2->tmpcost;
	
	/* update the xsep of the blocks following the swaped items */
	deltaxdim = tmpobj1->module->xdim - tmpobj2->module->xdim;
	if ( deltaxdim != 0 ) {
	    if ( tmpobj1->module->next != NULL ) {
		tmpobj1->module->next->xsep -= deltaxdim;
		tmpobj1->module->next->xtrans += deltaxdim;
	    }
	    if ( tmpobj2->module->next != NULL ) {
		tmpobj2->module->next->xsep += deltaxdim;
		tmpobj2->module->next->xtrans -= deltaxdim;
	    }
	}
    }
    
    /* displace the object to the right by displacement (could be negative) */
    else if ( movetype == DISPLACE ) {
	tmpobj1->module->xsep  += displacement;
	
	if ( tmpobj1->module->next != NULL ) {
	    tmpobj1->module->next->xsep   -= displacement;
	    tmpobj1->module->next->xtrans += displacement;
	}
    } else {
	tmpobj1->module->transform = (tmpobj1->module->transform == YES) ? NO : YES;
    }
}

/* This is the actual simulated annealing algorithm */
static void
  PERMsimAnneal(Temp0,Tempf,alpha,inloopiter,MODTYPE)
double Temp0,Tempf;
double alpha;
int inloopiter;
int MODTYPE;
{
    double Temp;
    double oldcost, newcost, prevcost;
    double avcost; /* median cost in the inner loop */
    double PERMinitCost();
    double ptime(), btime, etime;

    int loc1, loc2, movetype, displacement;

    int innerstop;
    int iternum;
    int num_unchanged;
    
    int acceptedmoves, swapcount, displacecount, mircount;
    int totswaps, totdis, totmirs;


    iternum = 0; totswaps = 0; totdis = 0; totmirs = 0;

    btime = ptime();

    avcost = oldcost = PERMinitCost();
	oldoverlap = 0;
	if ( Temp0 == 0 ) {
		Temp0 = 20 * maxxdim;
	}
    Temp = Temp0;
    num_unchanged = 0; /* number of consecutive times the cost is unchanged */
    while ( 1 ) {
	
        innerstop = 1;
        acceptedmoves = 0; swapcount = 0; mircount = 0; displacecount = 0;
        numcostinc = 0;
	
	prevcost = oldcost; /* cost before the innerloop iteration */
	avcost = 0;
        while ( innerstop <= inloopiter ) {
            PERMgenMove(&movetype, &loc1, &loc2, &displacement);
	    
	    newoverlap = oldoverlap;
	    /* PERMincremCost changes the value of newoverlap */
            newcost = oldcost + PERMincremCost( movetype, loc1, 
					       loc2, displacement );
	    
            if (PERMaccept(newcost,oldcost,Temp) == YES ) {
                PERMmakeMove(movetype, loc1, loc2, displacement, MODTYPE);
                acceptedmoves++;
                switch ( movetype ) {
		  case SWAP :
		    swapcount++; break;
		  case DISPLACE :
		    displacecount++; break;
		  case MIRROR :
		    mircount++; break;
                }
                oldcost = newcost;
		oldoverlap = newoverlap;
            }
	    avcost += oldcost;
            innerstop++;
        }
	avcost /= inloopiter; /* average cost in the inner loop */
	
        totswaps += swapcount;
        totdis   += displacecount;
        totmirs  += mircount;
	
        Temp *= alpha;
	
	if ( prevcost == oldcost ) {
	    num_unchanged++;
	} else {
	    num_unchanged = 0;
	}
	
        if ( Temp < Tempf || num_unchanged > 50 ) {
	    
            etime = ptime();
            break;
        }
        iternum++;
    }
}

static int
  PERMaccept(newcost,oldcost,Temp)
double newcost, oldcost;
double Temp;
{
    double deltacost;
    double overlapcost;
    double tempvar, randvar;
    
    if ( newoverlap == 0. ) {
	overlapcost = 0.;
    }
    else if ( newoverlap / Temp >= 80. ) {
	overlapcost = 5.5e34;
    }
    else {
	overlapcost = exp( newoverlap / Temp ) - 1;
    }
    deltacost = newcost - oldcost + overlapcost;
    if ( deltacost <= 0. ) { return( YES ); }
    else {
        tempvar = exp( - deltacost / Temp );
        randvar = PERMmakeRand();
        if ( randvar <= tempvar ) {
            numcostinc++; return ( YES );
        }
        else { return ( NO ); }
    }
}

static void
  PERMresOvLaps()
/* Resolve all the overlaps */
{
    int modcnt;
    struct _module *tmpmod;
    
    for ( modcnt = 1; modcnt <= nummod; modcnt++ ) {
	tmpmod = permlist[modcnt].module;
	if ( tmpmod->xsep < 0 ) {
	    if ( tmpmod->next != NULL ) {
		tmpmod->next->xsep += tmpmod->xsep;
		tmpmod->next->xtrans -= tmpmod->xsep;
	    }
	    tmpmod->xsep = 0;
	}
    }
}

static int
  PERMnetLimits(terminal, flag) 
struct _terminal *terminal;
int flag; /* EXACT or APPROX */
/* find the net_lowxlim and net_hixlim of a given terminal */
{
    int lowxcoord, hixcoord; 
    int tmpxcrd;
    struct _terminal *tmpterm;
    struct _eqterm *tmpeqt;

    if ( terminal->netptr->type == GROUND ||
                                    terminal->netptr->type == CLOCK ) {
        /*** grounded terminal */
        return ( GROUND );
    }

    terminal->net_lowxlim = MAXINT;
    terminal->net_hixlim = - (MAXINT - 1);

    /* for each terminal, find the module it is connecting to */
    tmpterm = terminal->netptr->firstnett;
    for ( ; tmpterm != NULL; tmpterm = tmpterm->nxtnett ) {
        if ( tmpterm != terminal && tmpterm->channel == terminal->channel &&
            tmpterm->module->type != SWCH  && tmpterm->module->type != OPAMP) { 
            if ( flag == EXACT ) {
                if ( (tmpeqt = tmpterm->eqterms) != NULL ) {
                    lowxcoord = MAXINT;
                    hixcoord = -(MAXINT - 1);
                    for ( ; tmpeqt != NULL; tmpeqt = tmpeqt->next ) {
                        tmpxcrd = PERMfindXeqt(tmpeqt);
                        if ( lowxcoord > tmpxcrd ) {
                            lowxcoord = tmpxcrd;
                        }
                        if ( hixcoord < tmpxcrd ) {
                            hixcoord = tmpxcrd;
                        }
                    }
                }
                else {
                    lowxcoord = hixcoord = PERMfindXcrd( tmpterm );
                }
            }
            else if ( flag == APPROX ) {
                lowxcoord = tmpterm->module->xtrans + tmpterm->module->xsep;
                hixcoord = tmpterm->module->xtrans +
                                tmpterm->module->xsep + tmpterm->module->xdim;
            }
            if ( lowxcoord < terminal->net_lowxlim ) {
                terminal->net_lowxlim = lowxcoord;
            }
            if ( hixcoord > terminal->net_hixlim ) {
                terminal->net_hixlim = hixcoord;
            }
        }
    }
    return( terminal->netptr->type );
}

static int
  PERMsingleCost(lastobj, xtrans, transform)
struct _permobj *lastobj; /* the object used for cost evaluation */
int xtrans; /* xtranslation of the object */
int transform; /* transformation of the object */
/* evaluate the cost of placing lastobj such that its left edge is at xtrans */ 
{
    int cost;
    int counter;
    int numterms;
    int lolimit, hilimit, trial; /* used for search in ordered list */


    numterms = lastobj->numterms;

    /* find where xtrans is located with respect to break points */
    if ( xtrans < lastobj->limitlist[transform][0] ) {
        lolimit = -1;
        hilimit = 0;
    }
    else if ( xtrans > lastobj->limitlist[transform][2 * numterms - 1] ) {
        lolimit = 2 * numterms - 1;
        hilimit = 2 * numterms;
    }
    else {
        lolimit = 0;
        hilimit = 2 * numterms - 1;
        while ( (hilimit - lolimit) > 1 ) {
            trial = (hilimit + lolimit) / 2;
            if ( xtrans < lastobj->limitlist[transform][trial] ) {
                hilimit  = trial;
            }
            else if ( xtrans > lastobj->limitlist[transform][trial] ) {
                lolimit = trial;
            }
            else {
                lolimit = hilimit = trial;
            }
        }
    }
    /* hilimit and lolimit differ by at most 1 */
    if ( hilimit <= numterms - 1 ) {
        /* the cost is sum of 1) slope times distance to 
                     closest rhs break point (hilimit) plus 2) cost at hilimit*/
        cost = (numterms - 1 - lolimit) *
                ( lastobj->limitlist[transform][hilimit] - xtrans ) +
                        lastobj->fvalue[transform][hilimit];
    }
    else if ( lolimit >= numterms ) {
        /* the cost is sum of 1) slope times distance to 
                     closest lhs break point (lolimit) plus 2) cost at lolimit*/
        cost = ( hilimit - numterms ) *
                ( xtrans - lastobj->limitlist[transform][lolimit]) +
                        lastobj->fvalue[transform][lolimit];
    }
    else {
		/* cost of the flat part of the function */
		/* we could use cost of lowlimit as well */
        cost = lastobj->fvalue[transform][hilimit];
    }
    return( cost );
}

static double
  PERMinitCost()
/* evaluate initial cost */
{
    double totcost;
    int xtrans, transform;
    int modcnt;
    struct _module *tmpmod;
    

	totcost = 0;
    for ( modcnt = 1; modcnt <= nummod; modcnt++ ) {
        tmpmod = permlist[modcnt].module;
        xtrans = tmpmod->xtrans + tmpmod->xsep; 
        transform = tmpmod->transform;
        permlist[modcnt].cost = 
                    PERMsingleCost(permlist + modcnt, xtrans, transform);
        totcost += permlist[modcnt].cost;
    }
    return(totcost);
}

static int
  PERMincremCost(movetype, object1, object2, displacement)
int movetype, object1, object2;
int displacement;
{
    int xtrans; 
    int oldxsep, newxsep; 
    struct _permobj *tmpobj1, *tmpobj2;
    
    
    tmpobj1 = permlist + object1;
    if ( movetype == SWAP ) {
        tmpobj2 = permlist + object2;
	
        xtrans = tmpobj2->module->xtrans + tmpobj2->module->xsep;
        tmpobj1->tmpcost = PERMsingleCost(tmpobj1, xtrans, 
					  tmpobj1->module->transform);
        xtrans = tmpobj1->module->xtrans + tmpobj1->module->xsep;
        tmpobj2->tmpcost = PERMsingleCost(tmpobj2, xtrans, 
					  tmpobj2->module->transform);
	
	/* total overlap or outside the bounds */
	if ( tmpobj1->module->xdim != tmpobj2->module->xdim ) {
	    if ( tmpobj2->module->next == NULL ) {
		oldxsep = maxxdim - tmpobj2->module->xtrans -
		  tmpobj2->module->xdim - tmpobj2->module->xsep;
		newxsep = maxxdim - tmpobj2->module->xtrans -
		  tmpobj1->module->xdim - tmpobj2->module->xsep;
	    }
	    else {
		oldxsep = tmpobj2->module->next->xsep;
		newxsep = tmpobj2->module->next->xsep - tmpobj1->module->xdim 
		  + tmpobj2->module->xdim;
	    }
	    if ( oldxsep >= 0 ) {
		if ( newxsep < 0 ) { newoverlap += - newxsep; }
	    }
	    else {
		if ( newxsep >= 0 ) { newoverlap += oldxsep; }
		else { newoverlap += oldxsep - newxsep; }
	    }
	    if ( tmpobj1->module->next == NULL ) {
		oldxsep = maxxdim - tmpobj1->module->xtrans -
		  tmpobj1->module->xdim - tmpobj1->module->xsep;
		newxsep = maxxdim - tmpobj1->module->xtrans -
		  tmpobj2->module->xdim - tmpobj1->module->xsep;
	    }
	    else {
		oldxsep = tmpobj1->module->next->xsep;
		newxsep = tmpobj1->module->next->xsep - tmpobj2->module->xdim 
		  + tmpobj1->module->xdim;
	    }
	    if ( oldxsep >= 0 ) {
		if ( newxsep < 0 ) { newoverlap += - newxsep; }
	    }
	    else {
		if ( newxsep >= 0 ) { newoverlap += oldxsep; }
		else { newoverlap += oldxsep - newxsep; }
	    }
	}
	
        return ( tmpobj1->tmpcost + tmpobj2->tmpcost -
		tmpobj1->cost - tmpobj2->cost);
    }
    else if (movetype == DISPLACE) {
        xtrans = tmpobj1->module->xtrans + tmpobj1->module->xsep + displacement;
        tmpobj1->tmpcost = PERMsingleCost(tmpobj1, xtrans, 
					  tmpobj1->module->transform);
	if ( tmpobj1->module->next == NULL ) {
	    oldxsep = maxxdim - tmpobj1->module->xtrans -
	      tmpobj1->module->xdim - tmpobj1->module->xsep;
	    newxsep = oldxsep - displacement;
	} else {
	    oldxsep = tmpobj1->module->next->xsep;
	    newxsep = oldxsep - displacement;
	}
	if ( oldxsep >= 0 ) {
	    if ( newxsep < 0 ) { newoverlap += - newxsep; }
	} else {
	    if ( newxsep >= 0 ) { newoverlap += oldxsep; }
	    else { newoverlap += oldxsep - newxsep; }
	}
	
	oldxsep = tmpobj1->module->xsep;
	newxsep = oldxsep + displacement;
	if ( oldxsep >= 0 ) {
	    if ( newxsep < 0 ) { newoverlap += - newxsep; }
	}
	else {
	    if ( newxsep >= 0 ) { newoverlap += oldxsep; }
	    else { newoverlap += oldxsep - newxsep; }
	}
	
        return ( tmpobj1->tmpcost - tmpobj1->cost );
    }
    else { 
        /* move is MIRROR */
        xtrans = tmpobj1->module->xtrans + tmpobj1->module->xsep;
        if ( tmpobj1->module->transform == YES ) {
            tmpobj1->tmpcost = PERMsingleCost(tmpobj1, xtrans, NO);
        }
        else {
            tmpobj1->tmpcost = PERMsingleCost(tmpobj1, xtrans, YES);
        }
        return ( tmpobj1->tmpcost - tmpobj1->cost );
    }
}

PERMmodules( MODTYPE, flag)
int MODTYPE; /* OPAMP or SWCH */
int flag; /* EXACT or APPROX */
{
    double Tzero, Tfinal;
    double alpha;
    int inloopiter;
    FILE *fp;


    PERMmSsgInit( MODTYPE, flag  );

    if((fp=fopen("SA.param","r")) == NULL) {
	Tzero = 0;
	Tfinal = 1.e-18;
	alpha = .9;
	inloopiter = 25 * nummod;
    } else {
	fscanf(fp,"%lf",&Tzero);
	fscanf(fp,"%lf",&Tfinal);
	fscanf(fp,"%lf",&alpha);    /* the value of alpha is between 0 and 1 */
	fscanf(fp,"%d",&inloopiter);
	if ( inloopiter <= 0 )     inloopiter = 25 * nummod;
	inloopiter *= nummod;
	fclose(fp);
    }
    
    seed = .31415926535897932384;
    PERMsimAnneal(Tzero,Tfinal,alpha,inloopiter,MODTYPE);
    
    PERMresOvLaps();
}
